///////////////
// SideBar.c //
///////////////
//
// Read copy of data stream going from TREMOR "prototype 0" instrument
// to CDPD modem and generate a bar-graph display for triggered
// earthquakes.
//
// Intended use is to allow instrument hosts to have immediate access
// to the last several days worth of earhquakes, and to let them see and
// get some perspective of PGA and PGV values experienced at their site.
//
// FILES:  All are on "tremor.wr.usgs.gov" in Menlo Park, CA, et seq.
//     Software is in directory "C:\jrevans\SideBar".  Configuration
//     file must also be IN ROOT DIRECTORY (C:\SideBarConfig.txt) and
//     it is ESSENTIAL that the sensor constants in it be correct for
//     the instrument being monitored.
//
// HISTORY:
//
//     V1.0:  18 November 1999 ... 28 September 2001:
//          Initial design.  Tabled for a long interval due to CRADA work.
//            Restarted August 2001.  Update to new metrics (PGV, Sa).
//            First working version available 28 September 2001.
//
//     V1.1:  16 October 2001:
//          Add PGV scale option for in/s.  Add display of number of
//            earhquakes in database.  Add button to jump to earliest
//            earthquake.  Increase size of Date/Time display text.
//            Improve orange color.
//
//     V1.2:  24 October 2001:
//          Add "Quit" button to make it harder to accidentally kill the
//            database by exiting.  Disabled "X" button.  Quits must now
//            be verified.
//          Add messages and warnings window to help with debugging.
//            Move many error messages to this window.
//          Finally found and corrected a destructive bug in FIFOgets()
//            (had "==" for "=", which mightilly confused the compiler).
//          RELEASE the binary of this version via Open-file Report (the
//            source is too sensitive to release under the GeoSIG CRADA).
//
//     V2.0:  08 November 2001:
//          Fix "==" bug in FIFOgets(), and add sanity checs and guick
//			  returns to FIFOstuff() and FIFOgets().
//
//     V3.0:  03 March 2003:
//          CHANGE TO A SINGLE BAR GRAPH OF INSTRUMENTAL INTENSITY,
//            shown in Roman numeral form, and numerical displays
//			  of PGA and PGV (in/s or cm/s).
//          Change color scale to "continuous tone" ShakeMap MMI.
//			  Smooth the scale bar accordingly.
//          Add history file to preserve database during shutdown
//            and after each new earthquake arrives.
//          Add Perceived Shaking and Potential Damage indications
//            following ShakeMap standards (minor editorial changes).
//
// FUTURE:
//     Get feedback from KTVU staff as a represenative media viewpoint.
//     Provide on-air/printable graphic accordingly.
//
//     Teach it to read waveforms and recompute metrics.  Show also the
//     horizontal and vertical maxima.  Add (verified) means to delete
//     false alarms?
//
// BUGS and PERVERSIONS:  Search all source code for "caveat", "caution",
//     "warning", and "bug" using case-insensitive and partial-word
//     settings.
//
//     NOTE in particular:  USER MUST KEEP THE PC'S CLOCK.  This clock
//     is the ONE AND ONLY source of timing for identifying the earthquake.
//     Try to keep it WITHIN A FEW SECONDS of correct local time.  Also
//     set TimeFudge to be the approximate time (in seconds) between the
//     onset of shaking and the "TRG0" line, indicating the start of a
//     trigger message, TimeFudge is usually about 90 s for the TREMOR
//     "prototype 0" instrument at maximum record length.
//
// J.R. Evans, USGS.

#include "SideBar.h"

#define STRICT
#include <windows.h>

HINSTANCE	hSaveCrntInst;	// handle to current instance

int WINAPI WinMain(
				   HINSTANCE	hCrntInst,
				   HINSTANCE	hPrevInst,  // always ==0 in Win32
				   LPSTR		lpszCmdArgs,
				   int			nCmdShow
				   )
{
	HWND		hWndMain;
	MSG			message;
	WNDCLASSEX	WndClass;

	// Save instance
	hSaveCrntInst = hCrntInst;

	// Fill in window class structure
	WndClass.cbSize			= sizeof(WNDCLASSEX);
	WndClass.style			= CS_HREDRAW | CS_VREDRAW | CS_NOCLOSE;
	WndClass.lpfnWndProc	= WndProc;
	WndClass.cbClsExtra		= 0;
	WndClass.cbWndExtra		= 0;
	WndClass.hInstance		= hCrntInst;
	WndClass.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	WndClass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);
	WndClass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	WndClass.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
	WndClass.lpszMenuName	= NULL;
	WndClass.lpszClassName	= WNDCLASSNAME;

	// Register the window class or return error
	if (0 == RegisterClassEx(&WndClass))
		return (TSERR_NoWndClassRegd);

	// Create top-level window
	hWndMain = CreateWindow(
		WNDCLASSNAME,			// Window-class Name
		APPNAME,				// Title bar text
		WS_OVERLAPPEDWINDOW,	// Window style
		0,						// x position (right)
		0,						// y position (down)
		CW_USEDEFAULT,			// x size (width)
		CW_USEDEFAULT,			// y size (height)
		NULL,					// Parent (NULL==top-level windows have none)
		NULL,					// Menu
		hCrntInst,				// Window owned by WinMain instance
		NULL					// Pointer to window-creation data
		);
	if (NULL == hWndMain)
		return (TSERR_NoWndCreated);

	// Make window visible and force update of its client area
	ShowWindow(hWndMain, nCmdShow);	// Show it
	UpdateWindow(hWndMain);			// Force a WM_PAINT message

	// Grand message-handling loop (loop exits on WM_QUIT)
	while (GetMessage(&message, NULL, 0, 0)) {	// NULL==any window, 0,0==next message
		TranslateMessage(&message);	// Translate virtual keys (e.g. shift)
		DispatchMessage(&message);	// Send message to WndProc()
	}

	return(message.wParam);
}
