///////////////////////
// TREMOR_Receiver.h //
///////////////////////

#ifndef TREMOR_RECEIVER_H
#define TREMOR_RECEIVER_H

// Includes ///////////////////////////////////////////////////////////////
#include <stdio.h>		// Includes define of "NULL"

#define STRICT
#include <windows.h>

#include <string.h>		// Includes define of "NULL"
#include <math.h>

#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>

// BOOL, TRUE, and FALSE are defined in "windef.h",
// which is included via "windows.h".

#include "datetime.h"
#include "rsc_result.h"

// Defines ////////////////////////////////////////////////////////////////
#define APPNAME			"TREMOR_Receiver, V5.0, 31 July 2001"
#define WNDCLASSNAME	"C_TREMOR_Receiver"

// Custom Windows messages
#define WM_UDP_RECV		(WM_USER + 1101)
#define WM_PEELALL		(WM_USER + 1102)
#define WM_PEELTRG		(WM_USER + 1103)
#define WM_PEELEVT		(WM_USER + 1104)
#define WM_DSPLYRCNT	(WM_USER + 1105)
#define WM_DSPLYWARN	(WM_USER + 1106)


// IDs for WM_COMMAND messages (unique IDs for various controls)
// Menu IDs
#define IDM_BASE		(100)

// Button IDs
#define IDB_BASE		(200)
#define IDB_QUIT		(IDB_BASE + 1)
#define IDB_NEWLOG		(IDB_BASE + 2)
#define IDB_CLRWRN		(IDB_BASE + 3)
#define IDB_INTERRUPT	(IDB_BASE + 4)
#define IDB_STROBE		(IDB_BASE + 5)

// List-box IDs
#define IDLB_BASE		(300)
#define IDLB_SITE		(IDLB_BASE + 1)

// Edit-window IDs
#define IDE_BASE		(400)
#define IDE_EDIT		(IDE_BASE + 1)
#define IDE_WARN		(IDE_BASE + 2)
#define IDE_TYPE		(IDE_BASE + 3)

// Button labels for UDPtalk() dis/enable.
#define START_SITE_TALK	"Push to Write"
#define END_SITE_TALK	"To Quit Write w/"

// Misc magic numbers
#define CONTROL_WIDTH	((int)20)
					// Width of control strip, in characters
#define SITE_LIST_HEIGHT ((int)300)
					// site listbox height
#define EDIT_HEIGHT_DEF	((int)36)
					// Lines visible at one time (default)
#define EDIT_WIDTH		((int)83)
					// Characters visible across line (try 60 or 83)
#define DSPLY_BUF_SIZE	((int)30000)
					// DisplayQ[] size.
					// Try EDIT_WIDTH * EDIT_HEIGHT_DEF * (15 pages)

#define BUF_SIZE		((int)512)
					// General character buffer size
#define DIR_NAME_SIZE	((int)256)
#define SHORT_STR		((int)16)
#define MEDIUM_STR		((int)24)
#define LONG_STR		((int)32)

#define smp_per_pkt		((int)100)
					// Number of samples expected in a single
					// waveform packet (excepting end packets)

#define MAX_RAVEN_UDP	((int)600)
					// Biggest UDP packet a Raven modem can send
#define UDP_BUF_SIZE	((int)1500)
					// UDP I/O buffer size (must be >= MAX_RAVEN_UDP + 1
					// + number of lines).  The last byte of the UDP
					// buffer is reserved for an added null byte, since
					// it is digested as a string.  The additional byte
					// per line is for expansion of CR to CR-NL for PC.
#define VER_BUF_SIZE	((int)6000)
					// Maximum verbose-packet size.  Try (4 * UDP_BUF_SIZE).
#define MAX_INTRANSIT_BUFFERS ((int)50)
					// TREMOR_Receiver() processes UDP packets as it gets
					// them, and may be processing many at one time.  This
					// parameter is the largest number of packets that can
					// be processed simultaneously.  It limits allocation
					// if buffer memory by UDPsocket().  TREMOR_Receiver()
					// WILL STOP if this limit is exceeded, but may burn
					// up too much memory if MAX_INTRANSIT_BUFFERS is set
					// too high--adjust with care.
#define WARN_BUF_SIZE	((int)1024)
					// Warnings buffer size for each site.
#define VERIFY_REPEATS	((int)3)
					// Repitions of verification packets from
					// TREMOR_Receiver() to the TREMOR site.
					// Also, repititions of site-interrupt signal.

// Ensure that MAX_SENSORS >= 3 * MAX_UNITS and that MAX_UNITS >= MAX_SITES.
#define MAX_SITES		((int)15)
					// Maximum number of TREMOR sites supported
#define MAX_UNITS		((int)25)
					// Maximum number of TREMOR recorders supported
#define MAX_SENSORS		((int)100)
					// Maximum number of ICS accels supported (must also be
					// large enough to read in all of "MfgCalibrations.txt".
#define MAXRECIPIENTS	((int)10)
					// Maximum number of real-time-message recipients
#define MAXCLKERR		((int)10)
					// Length of clock-error history (number of entries).
#define MAXTRIGTIMES	((int)100)
					// Number of recent trigger times to remember.
					// Used to fix waveform arrival times.
#define MAX_REPROCESS	((int)100)
					// Maximum number of files that can be waiting
					// to have their shaking metrics recomputed.
					// Over this, metrics are just not recomputed.

#define CLKUDPLAG		((double)0.294)
					// Estimated transit time for UDP packets, from
					// field instrument to TREMOR_Receiver.  (s)
					// Not currently used.

// Some averages of 20, discarding round trip times above 650 ms to
// estimate a smoothed minimum (values shown are 1/2 of the two-way
// traveltime reported by PingPlotter().).  03 Nov 1999:
// P02  0.294 s
// P03  0.295 s
// P04  0.289 s
// P05  0.280 s
// P06  0.303 s
// P07  0.295 s
// P10  0.297 s
// P11  0.298 s
// P12  0.295 s  +/- 0.010
// -----------------------
// Mean 0.294 s  +/- 0.006

#define VERTICAL		((int)1)
#define NORTH			((int)2)
#define EAST			((int)3)
#define TEMPERATURE		((int)4)
					// Channel ID in PeelEvt()

// Timers
#define TIMER_KICK		((unsigned int)1)		// ID; log-file restarts
#define PERIODIC_KICK	((unsigned int)64010)	// ms
#define TIMER_UDP		((unsigned int)2)		// ID; UDP strobe
#define UDP_KICK		((unsigned int)7520)	// ms
#define TIMER_HEART		((unsigned int)3)		// ID; heartbeats
#define HEARTBEAT		((unsigned int)128070)	// ms

#define NULL_WAVE_P		((short int)(32767))
#define NULL_WAVE_M		((short int)(-32767))
					// Waveform magic-marker numbers

// Numerical constants
#define ADCtoV			((double)10. / (double)4095.)
					// * to convert ADC counts to Volts
#define ONE_G			((double)980.665)
					// gal, cm/s^2, mean surface gravity
#define ZIP				((double)0.0)
#define HALF			((double)0.5)
#define ONE				((double)1.0)
#define SIXTEEN			((double)16.0)
#define HH				((double)100.)
#define KK				((double)1000.)

// #define HUGE_IEEE			((double)1.7976931348623158e308)
					// Biggest 64-bit IEEE double
#define INF				HUGE_VAL				// "Infinity"
#define PI				((double)3.14159265358979324)
#define TWOPI			((double)2.0 * PI)


// Thermistor constants
#define R0				((double)49800.0)		// Ohms in series with thermistor
#define Vref			((double)2.50)			// Reference voltage across series pair
#define AAA				((double)1.034034113e-03)// Coefficients of cubic
#define BBB				((double)2.383800240e-04)
#define DDD				((double)1.592455177e-07)// CCC=0

#define EPOCH_ID		"%#!SendersEpoch1970="
					// String uniquely identifying an epochal
					// time listed within the UDP packet that
					// should be used to name the output file.
#define FILENAME_ID		"%#!SendersFileName="
					// String giving the sender's filename,
					// which will be used as the UDP output
					// filename.
#define HEARTBEAT_ID	"%#!HeartbeatFileName="
					// Marks a heartbeat file intended to be
					// overwritten at the receiving end and
					// having the fixed filename indicated.
#define FILENAME_END	"%#!; \t\r\n\0"
					// Terminates the filename indicated by
					// HEARTBEAT_ID or FILENAME_ID.
#define HEARTBEAT_FN	"AllsWell.hrt"
					// Fixed filename for heartbeats.
#define CMNCTV_STA_WNDW	24
					// Window (hours) during which station must be
					// heard from to be considered "still alive".
#define VerboseLB		"\r\n          "
					// Used by Verbose() for line breaks

// Define known Sa periods and corresponding annotations //////////////////
// (MUST correspond to what ft() sends.)
#define SaPer1			( 300)		// ms
#define SaPer2			(1000)		// ms
#define SaPer3			(3000)		// ms

#define SaNote1			"Sa#1"
#define SaNote2			"Sa#2"
#define SaNote3			"Sa#3"

#define OLD_METRICS		(1)
#define NEW_METRICS		(2)

// Tolerance ranges, beyond which an incoming value is highlighted ////////
#define MIN_BAT			((double)12.0)			// VDC
#define MAX_BAT			((double)14.0)
#define MIN_TEMP		((double)-5.0)			// degrees C
#define MAX_TEMP		((double)45.0)
#define MAX_NOISE		((double)10.0)			// cm/s/s
#define MAX_SLOW		((double)-2.0)			// seconds
#define MAX_FAST		((double)2.0)			// (FAST is POSITIVE)

// Macros /////////////////////////////////////////////////////////////////
#define MIN(a, b)		((a) < (b) ? (a) : (b))
#define MAX(a, b)		((a) > (b) ? (a) : (b))
#define MAX3(a, b, c)	(MAX(MAX(a, b), c))
#define streq			!strcmp
#define strneq			!strncmp
#define VERBOSE_ERR(nt)	{if (nt != ntoks) {ErrWarn("\r\nWrong number of tokens in Verbose().\r\n", ist); sprintf(VerboseTextOut, "%s   ???", RawTextIn);} }

// Errors returned ////////////////////////////////////////////////////////
#define TSERR_NoWndClassRegd	1
#define TSERR_NoWndCreated		2

///////////////////////////////////////////////////////////////////////////
// Function prototypes.  This is their order in file "TR_implement.cpp"
// excepting those indicated in other source files.

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
// then alphabetic
void		CheckMSGSTR(char *, int);
void		CheckVerStrLen(char *, int, int);
void		ClockErrAdd(double, double, int);
double		ClockErrFix(double, double, int, char *, BOOL);
double		CompArias(long, double, double, double, double, double);
double		CompPGV(long, double, double, double, double, double);
double		CompPSV(long, double, double, double, double, double);
double		CompSa(long, double, double, double, double, double);
double		Compensate(int, int, double, short int);
void		ConfigReader();
void		Controls(HWND, int, int);
void		DsplyRcnt();
void		DsplyWarn();
void		EndConversation();
char *		EpochToStr(double);
void		ErrExit(char *);
void		ErrWarn(char *, int);
void		FrontVerbose(char *, char *, int, int);
void		list_trigger_errors(unsigned int, int);
void		MakeControls(HINSTANCE);
void		OpenTranscript(int);
void		PeelAll(WPARAM, LPARAM);
void		PeelEvt(WPARAM, LPARAM);
void		PeelTrg(WPARAM, LPARAM);
void		pgv(double *, int *, double, double *, int);
void		ReadSensorLists();
void		ReadSiteLists();
void		ReadTrace(FILE *, char *, double *, int);
void		RefreshSocket(int);
DWORD WINAPI ReviseMetrics();
rsc_result	rsc(double, double, double, double *, int);
void		StartUDPstrobe();
void		StoreTriggerTimes(double, double, int);
void		strchk(char *, int, char *, int);
char *		strspnp(char *, char *);	// My version since Mr. Bill's documentation lies.
double		Temperature(short int);
void		TestBatt(double, char *, int);
void		TestClock(double, double, int);
void		TestNoise(double, int);
void		TestTemp(double, char *, int);
void		UDPsocket(WPARAM, LPARAM);
BOOL		UDPstart();
void		UDPtalk(char, LPARAM);
void		UDPverify(int, char *);
double		VBat(short int);
void		Verbose(char *, char *, int, int);
void		write_wv_pkt(FILE *, char *, int, int);

// In l1fit.c
void		free_fvector(double *, int, int);
double *	fvector(int, int);
void		free_ivector(int *, int, int);
int *		ivector(int, int);
void		l1fit(int, int, double * *, double *, double, double *, double *);

///////////////////////////////////////////////////////////////////////////
// Structure typedefs

// Main data structure used by TREMOR_Receive()
typedef struct {
	char			TREMOR_SN[MEDIUM_STR];	// TREMOR unit at site

	double			latitude;				// Site location ...
	double			longitude;
	double			elevation;
	double			N_azimuth;				// ... and orientation of "North" ...
	double			N_azm_err;				// ... and its uncertainty (degrees) ...
	char			StationSite[LONG_STR];	// ... and site name/description.

	// "Fr" ("From") and "To" are the usual data transfer directions
	// (FROM site TO Menlo Park) but are a little perverse sounding
	// when sending instructions from Menlo to the site in UDPtalk().
	unsigned short	FrIP[4];				// Modem IP address ...
	unsigned short	FrPort;					// ... and port (i.e., the site)
	unsigned short	ToPort;					// Matching port of TREMOR_Receive()
											// (cf. "ToIP", above)

	char			V_ICS_SN[SHORT_STR];	// Sensor S/N (ICS)
	double			V_Sensitivity;			// Sensor sensitivity ("gain") at ~22.5 C (V/gal)
	char			N_ICS_SN[SHORT_STR];	//    .
	double			N_Sensitivity;			//    .
	char			E_ICS_SN[SHORT_STR];	//    .
	double			E_Sensitivity;			//    .
											// (Reference temperature should be
											// entered via cal files, and a
											// variable here.  22.5 C is nominal.)

	double			V_TCS, V_Tref;			// Gain temperature sensitivities
	double			N_TCS, N_Tref;			// (fraction/C) and respective
	double			E_TCS, E_Tref;			// reference temperatures (C).

	SOCKET			sk;						// local UDP socket
	SOCKADDR_IN		addr_there;				// address info for TREMOR site
	SOCKADDR_IN		addr_here;				// address info for local socket

	int				LastPktSize;			// Used to identify redundant packets
	int				LastPktID;				// ==0 if "none" or "unknown", so ft()
											// must NEVER issue a packet number 0.

	char			DisplayQ[DSPLY_BUF_SIZE]; // Buffer used by DsplyRcnt()
	int				DQsize;					// Number of characters already in
											// DisplayQ (less terminating null).

	char			Warn[WARN_BUF_SIZE];	// Text buffer for warnings issued
											// about this site's data stream.
	int				WarnSize;				// Number of characters in it now
											// (less terminating null).

	FILE			*fp_all;				// Pointer to transcript file stream.
	char			fn_all[BUF_SIZE];		// Corresponding file name.

	FILE			*fp_evt;				// File pointers for Event output ...
	char			fname[DIR_NAME_SIZE];	// ... and the associated filename.

	double			PktArrTime;				// Aproximate system time that the
											// most recent UDP packet arrived
											// for that site.

	// The following stored oldest at [0]:
	int				nClockErrors;			// Number of data in error history.
	double			ClockError[MAXCLKERR];	// Recent history of clock errors.
	double			ClockErrTime[MAXCLKERR];// WHEN these errors were measured.

	// The following used to remember trigger times and their corresponding
	// corrected times.  These are used to correct subsequent waveform times.
	// Done this way because it may not be possible to back-project an
	// accurate correction at the time the waveforms are received.
	// The following stored oldest at [0]:
	int				nTriggerTimes;
	double			RawTriggerTime[MAXTRIGTIMES];
	double			TrueTriggerTime[MAXTRIGTIMES];

	// The last time the station communicated with TREMOR_Receiver().
	// That is, the last time that a begin, trigger, SOH, or waveform
	// message was received (PC time base).
	double			LastKnownMessage;
} SiteData;

////////////////////////////////////////////////////////////////////////////////
// Preliminary data sets from which "SiteData" is built
typedef struct {
	char			ModemType[SHORT_STR];
	char			ModemModel[LONG_STR];
	char			ModemSN[LONG_STR];
	char			ModemEID[LONG_STR];
	unsigned short	ModemIP[4];
	unsigned short	ModemPort;
	unsigned short	ToPort;
	char			TREMOR_SN[MEDIUM_STR];
} Communications;

typedef struct {
	char			TREMOR_SN[MEDIUM_STR];
	double			latitude;
	double			longitude;
	double			elevation;
	double			N_azimuth;
	double			N_azm_err;
	char			StationSite[LONG_STR];
} Coordinates;

typedef struct {
	char	TREMOR_SN[MEDIUM_STR];
	char	V_SN[SHORT_STR], V_ICS_SN[SHORT_STR], V_Comment[SHORT_STR];
	double	V_Sensitivity;			// Sensor sensitivity at room T (V/gal)
	double	V_Tref;					// Temperature to which this sensitivity applies
	char	N_SN[SHORT_STR], N_ICS_SN[SHORT_STR], N_Comment[SHORT_STR];
	double	N_Sensitivity;
	double	N_Tref;
	char	E_SN[SHORT_STR], E_ICS_SN[SHORT_STR], E_Comment[SHORT_STR];
	double	E_Sensitivity;
	double	E_Tref;
	char	TREMOR_SN_dup[MEDIUM_STR];
	double	VOrient[2], NOrient[2], EOrient[2];
} SerialNumbers;

typedef struct {
	char	ICS_Model[SHORT_STR];	// ICS Model
	char	ICS_SN[SHORT_STR];		// ICS Serial Number
	int		Proc;					// ICS Process Number
	double	Sens;					// Sensitivity (V/gal) {gal = cm/s^2}
	double	Mxg;					// Maximum linear range (gal)
	double	Offset;					// Zero offset (uncompensated) (gal)
	double	Bridge;					// Bridge resistance (Ohms)
	double	damping;				// Fraction of critical damping
	double	Hz;						// Natural frequency
	double	TCZ;					// Temperature coefficient of zero (gal/C)
	double	TCS;					// Temperature coefficient of Sensitivity (fraction/C)
	double	TCR;					// Temperature coefficient of Bridge R (Ohms/C)
	double	Rset;					// Compensating Resistor, TCS (Ohms)
	double	Rstc;					// Compensating Resistor, TCS (Ohms)
	double	Rztc1;					// Compensating Resistor, TCZ (Ohms)
	double	Rztc2;					// Compensating Resistor, TCZ (Ohms)
	double	Rs;						// Compensating Resistor, TCS (Ohms)
	double	Rzb1;					// Compensating Resistor, Offset (Ohms)
	double	Rzb2;					// Compensating Resistor, Offset (Ohms)
	double	FivePercentHz;			// 5% response frequency [Hz]; Inf if not present.
	char	CalComment[LONG_STR];	// Misc. comment on sensor
} Calibrations;

// Structure used by Verbose so stations don't trip
// over one another during translation, but do remember
// certain information from one UDP packet to another
// or even from one line to another.
typedef struct {
	double	sps;					// Sample rate
	BOOL	sps_well_known;			// TRUE when known from a BGN0, FALSE when assumed.
									// (CAVEAT:  set but not used -- should be used)
	BOOL	EventTimesOK;			// Whether PreEvent and TotalEvent are real.
	double	PreEvent;				// Pre-event duration (s)
	double	TotalEvent;				// Total event duration (s)
									//--------------------------------------
	double	TmprtrA;				// C (for use with Arias, PSV, PGV, and Sa)
	double	TmprtrW;				// C (used ONLY to correct waveforms)
									//--------------------------------------
	int		WaveCorrQual;			// 0 = uncorrected; 1 = fair correction;
									// 2 = corrected identically to trigger.
									//--------------------------------------
	double	soh_time;				// Corrected SOH arrival times.
	double	RawTriggerTime;
	double	TrueTriggerTime;
									//--------------------------------------
	double	wave_time;
	int		WaveLinesRemaining;		// Used by Verbose() to monitor the
									// progress toward the end of a waveform
									// UDP packet.  >0 causes translation of a
									// line which otherwise would be passed as
									// is because it would be "unrecognized".
	int		WaveLinesCh;			// Channel number of the UDP packet
	unsigned int	err_flgs;		// Computational-error flags from ft()
} VerSave;


#endif // TREMOR_RECEIVER_H
