///////////////////////////////////////////////////////////////////////////////
// L1FIT() -- L1 norm fitting
///////////////////////////////////////////////////////////////////////////////
// Fit "Ax = b" by minimum absolute residuals.
// Solution of an overdetermined system of equations in the L1 norm.
// Barrodale and Roberts, CACM (June 1974) pp 319-320.
// Algorithm 478

// Originally translated from the FORTRAN by Bruce Julian.

// Changed from float to double, and single-character variable names
// doubled, and FORTRAN-->C translation checked by JREvans, Nov 1999.

// Note that use of [1...] indexed arrays makes for a slightly messy
// interface to your calling routine.  One solution is to allocate
// aa[] as a zero-indexed ([0...]) array of pointers to doubles, and
// then to call l1fit(..., aa - 1, ...).  Typically, bb should be
// indexed the same as aa to avoid confusion in the calling routine,
// hence calling as l1fit(..., aa - 1, bb - 1, ...).  A successful
// example of a calling routine is appended below.

#include <stdio.h>
#include <math.h>
#include "local.h"

#include <stdlib.h>
#include "libq.h"

// B&R74 used 10^75 for IBM double precision
#define BIGCONST		((double)1.0e20)
#define SWAP(xx, yy)	{ dd = xx ; xx = yy ; yy = dd; }

	double	fabs();				// -lm

void
l1fit(mm, nn, aa, bb, toler, xx, ee)
	int		mm;					// Number of equations
	int		nn;					// Number of unknowns (mm >= nn)
	double	**aa;				// [1..mm+2][1..nn+2]
								// [1..mm][1..nn] = Coefs. (destroyed)
								// On exit,
								// aa[mm+1][nn+1] = Sum of |residuals|
								// aa[mm+1][nn+2] = Rank of coef. matrix
								// aa[mm+2][nn+1] = Exit code
								// aa[mm+2][nn+2] = No. of iterations done
	double	*bb;				// [1..mm] Right-hand side (destroyed)
	double	toler;				// Small positive tolerance.  (Try 10^-2d/3,
								// where d is the number of decimal digits
								// of accuracy.  Hence try 10^-10 for 64-bit
								// IEEE floats, and 10^-4 for 32-bit IEEE.)
	double	*xx;				// [1..nn] Solution
	double	*ee;				// [1..mm] Residuals
{
	double	pivot;
	double	dd;					// For SWAP() and other local, temporary use.
	int		kount;
	int		in;
	double	sum;
	double	min, max;
	int		out;
	int		stage;
	int		test;
	int		ii, jj;				// Dummy loop indices
	int		kk;
	int		kl, kr;
	int		ll;
	int		m1, n1;				// mm+1, nn+1
	int		m2, n2;				// mm+2, nn+2
	int		*ss;				// [1..mm] Work area

	// Initialization
	ss = ivector(1, mm);
	m1 = mm + 1;
	n1 = nn + 1;
	m2 = mm + 2;
	n2 = nn + 2;

	for (jj = 1 ; jj <= nn ; jj++) {
		aa[m2][jj] = (double)jj;
		xx[jj] = 0.0;
	}

	for (ii = 1 ; ii <= mm ; ii++) {
		aa[ii][n2] = (double)(nn + ii);
		aa[ii][n1] = bb[ii];
		if (bb[ii] < 0.0) {
			for (jj = 1 ; jj <= n2 ; jj++) {
				aa[ii][jj] = -aa[ii][jj];
			}
		}
		ee[ii] = 0.0;
	}

	// Compute the marginal costs
	for (jj = 1 ; jj <= n1 ; jj++) {
		sum = (double)0.0;
		for (ii = 1 ; ii <= mm ; ii++) {
			sum += aa[ii][jj];
		}
		aa[m1][jj] = (double)sum;
	}
	
	// Stage 1
	// Determine the vector to enter the basis
	stage = TRUE;
	kount = 0;
	kr = 1;
	kl = 1;
	FOREVER {	// ----------------------------------------------- B&R74 "70"
		max = -1.0;

		for (jj = kr ; jj <= nn ; jj++) {
			if (fabs(aa[m2][jj]) <= (double)nn) {
				dd = fabs(aa[m1][jj]);
				if (dd > max) {
					max = dd;
					in = jj;
				}
			}
		}

		if (aa[m1][in] < (double)0.0) {
			for (ii = 1 ; ii <= m2 ; ii++) {
				aa[ii][in] = -aa[ii][in];
			}
		}

		FOREVER {	// ------------------------------------------ B&R74 "100"

			// Determine the vector to leave the basis
			kk = 0;
			for (ii = kl ; ii <= mm ; ii++) {
				dd = aa[ii][in];
				if (dd > toler) {
					kk++;
					bb[kk] = aa[ii][n1] / dd;
					ss[kk] = ii;
					test = TRUE;
				}
			}
			
			FOREVER {	// -------------------------------------- B&R74 "120"
				if (kk <= 0)
					test = FALSE;
				else {
					min = BIGCONST;
					for (ii = 1 ; ii <= kk ; ii++) {
						if (bb[ii] < min) {
							jj = ii;
							min = bb[ii];
							out = ss[ii];
						}
					}

					bb[jj] = bb[kk];
					ss[jj] = ss[kk];
					kk--;
				}

				// Check for linear dependence in stage 1
				if (!test && stage) {
					for (ii = 1 ; ii <= m2 ; ii++)
						SWAP(aa[ii][kr], aa[ii][in]);
					kr++;
					goto BR260;
				}

				if (!test) {
					aa[m2][n1] = 2.0;
					goto BR350;
				}

				pivot = aa[out][in];
				if (aa[m1][in] - pivot - pivot <= toler)
					goto BR200;

				for (jj = kr ; jj <= n1 ; jj++) {
					dd = aa[out][jj];
					aa[m1][jj] -= (dd + dd);
					aa[out][jj] = -dd;
				}

				aa[out][n2] = -aa[out][n2];
			}

BR200:	// ------------------------------------------------------ B&R74 "200"

			// Pivot on aa[out][in]
			for (jj = kr ; jj <= n1 ; jj++) {
				if (jj != in)
					aa[out][jj] /= pivot;
			}

			for (ii = 1 ; ii <= m1 ; ii++) {
				if (ii != out) {
					dd = aa[ii][in];
					for (jj = kr ; jj <= n1 ; jj++) {
						if (jj != in)
							aa[ii][jj] -= dd * aa[out][jj];
					}
				}
			}

			for (ii = 1 ; ii <= m1 ; ii++) {
				if (ii != out)
					aa[ii][in] /= -pivot;
			}

			aa[out][in] = 1.0 / pivot;
			SWAP(aa[out][n2], aa[m2][in]);
			kount++;

			if (!stage)
				goto BR270;

			// Interchange rows in stage 1
			kl++;
			for (jj = kr ; jj <= n2 ; jj++) {
				SWAP(aa[out][jj], aa[kount][jj]);
			}

BR260:		if (kount + kr != n1)	// -------------------------- B&R74 "260"
				break;

			// Stage 2
			stage = FALSE;

			// Determine the vector to enter the basis
BR270:		max = -BIGCONST;	// ------------------------------ B&R74 "270"

			for (jj = kr ; jj <= nn ; jj++) {	// B&R74 "290" loop
				dd = aa[m1][jj];

				if (dd < (double)0.0) {
					if (dd > (double)(-2.0))
						continue;
					dd = -dd - (double)2.0;
				}

				if (dd > max) {
					max = dd;
					in = jj;
				}
			}

			if (max <= toler)
				goto BR310;

			if (aa[m1][in] <= 0.0) {
				for (ii = 1 ; ii <= m2 ; ii++)
					aa[ii][in] = -aa[ii][in];
				aa[m1][in] -= (double)2.0;
			}

		}	// B&R74 "100" loop end
	}	// B&R74 "70" loop end

BR310:

	// Prepare output
	ll = kl - 1;

	for (ii = 1 ; ii <= ll ; ii++) {
		if (aa[ii][n1] < 0.0) {
			for (jj = kr ; jj <= n2 ; jj++) {
				aa[ii][jj] = -aa[ii][jj];
			}
		}
	}

	aa[m2][n1] = 0.0;
	if (kr == 1) {
		for (jj = 1 ; jj <= nn ; jj++) {
			dd = (double)fabs(aa[m1][jj]);
			if (dd <= toler || ((double)2.0 - dd) <= toler)
				goto BR350;
		}
		aa[m2][n1] = 1.0;
	}

BR350:

	for (ii = 1 ; ii <= mm ; ii++) {
		kk = (int)aa[ii][n2];
		dd = aa[ii][n1];
		if (kk <= 0) {
			kk = -kk;
			dd = -dd;
		}

		if (ii < kl) {
			xx[kk] = dd;
		}
		else {
			kk -= nn;
			ee[kk] = dd;
		}
	}

	aa[m2][n2] = (double)kount;
	aa[m1][n2] = (double)(n1 - kr);

	sum = (double)0.0;
	for (ii = kl ; ii <= mm ; ii++) {
		sum += aa[ii][n1];
	}
	aa[m1][n1] = (double)sum;

	free_ivector(ss, 1, mm);
}


#if 0
///////////////////////////////////////////////////////////////////////////////
// Test example for l1fit().

#define CHUNKSIZE 10		// No. of rows allocated each time
#define u_int	unsigned

// Read in coordinates of some points, fit a straight line to them.

main()
{
	int		mm;
	int		nn;
	double	**aa;			// [1..mm+2][1..nn+2]
	double	*xx;			// [1..nn]
	double	*ee;			// [1..mm]
	double	*bb;			// [1..mm]
	double	xp, yp;			// Coordinates of a point to fit
	u_int	dim;			// Allocated rows in aa, bb

	nn = 2;
	mm = 0;

	// Read in data, allocating storage as needed
	dim = 0; aa = (double**)NULL; bb = (double*)NULL;
	while (scanf("%f%f", &xp, &yp) == 2) {
		while (dim < mm + 3) {
			// Need space:  Increase allocation of aa, bb
			dim += CHUNKSIZE;
#define GETMEM(xx, tt, dd)	(tt*)(xx == NULL ? emalloc(dd * sizeof(tt)) : erealloc((char*)aa, dd * sizeof(tt)))
			aa = GETMEM(aa, double*, dim);
			bb = GETMEM(bb, double , dim);
		}
		aa[mm] = fvector(1, nn+2);
		aa[mm][1] = xp; aa[mm][2] = 1.0; bb[mm] = yp;
		mm++;
	}
	aa[mm]   = fvector(1, nn+2);
	aa[mm+1] = fvector(1, nn+2);

	// Fit straight line to data, print result
	xx = fvector(1, nn);
	ee = fvector(1, mm);	// Bruce Julian had (1, nn) which looks like a bug.
	l1fit(mm, nn, aa-1, bb-1, 1.0e-10, xx, ee);
	(void)printf("return %g\n", (aa-1)[mm+2][nn+1]);
	(void)printf("yy = %g xx + %g\n", xx[1], xx[2]);

	// Free allocated storage
	free_fvector(ee, 1, mm);// Bruce Julian had (1, nn) which looks like a bug.
	free_fvector(xx, 1, nn);
	mm += 2;
	while (mm-- > 0)
		free_fvector(aa[mm], 1, nn+2);
	free((char *)aa);
	free((char *)bb);
}
#endif


///////////////////////////////////////////////////////////////////////////////
// Subroutines used by l1fit() (redundant with those in Julian's libq and
// stollen from there).

// Allocate an int vector with range [lwr..hir]
int *
ivector(lwr, hir)
	int		lwr, hir;
{
	return((int *)emalloc((unsigned)(hir - lwr + 1) * sizeof(int)) - lwr);
}

// Free an int vector allocated by ivector()
void
free_ivector(vv, lwr, hir)
	int		*vv;
	int		lwr, hir;
{
	free((char *)(vv + lwr));
}

// Allocate a double vector with range [lwr..hir]
double *
fvector(int lwr, int hir)
{
	return((double *)emalloc(
		(unsigned)(hir - lwr + 1) * sizeof(double)) - lwr
		);
}

// Free a double vector allocated by fvector()
void
free_fvector(vv, lwr, hir)
	double	*vv;
	int		lwr, hir;
{
	free((char *)(vv + lwr));
}

#if 0


///////////////////////////////////////////////////////////////////////////////
// A successful example:

///////////////////////////////////////////////////////////////////////////////
// Correct a clock back to TREMOR_Receiver()'s time base.  Returns corrected
// version of epochal time InstClockToFix, using an error predicted as of
// epochal time MasterClockNow.  PredictedError is POSITIVE if instrument
// clock is predicted to be FAST.

// ***** process *****
double
ClockErrFix(HWND hWndMain,
	double InstClockToFix, double MasterClockNow, int ist)
{
	double	PredictedError = (double)0;
	int		ii;

	// For l1fit():
	int		mm;						// Number of equations
	int		nn;						// Number of unknowns (mm >= nn)
	double	*aa[MAXCLKERR + 2];		// [1...mm+2][1...nn+2]
									// [1...mm][1...nn] = Coefs. (destroyed)
									// On exit,
									// a[mm+1][nn+1] = Sum of |residuals|
									// a[mm+1][nn+2] = Rank of coef. matrix
									// a[mm+2][nn+1] = Exit code (1=success;
									//      0=nonunique but OK; 2=failed)
									// a[mm+2][nn+2] = No. of iterations done
	double	bb[MAXCLKERR];			// [1...mm] Right-hand side (destroyed)
	double	toler = (double)1.0e-10;// Small positive tolerance.  (Try 10^-2d/3,
									// where d is the number of decimal digits
									// of accuracy.  So try 10^-10 for 64-bit
									// IEEE floats, and 10^-4 for 32-bit IEEE.)
	double	*xx;					// [1...nn] Solution
	double	*ee;					// [1...mm] Residuals

	// Predict current clock error for that instrument using L1 norm.
	mm = Sites[ist]->nClockErrors;
	nn = 2;

	if (mm <= 0) {
		// When no basis for correction, return unchanged instrument time.
		PredictedError = (double)0.0;
	}

	else if (mm == 1) {
		// When exactly one datum, use it to correct instrument time.
		PredictedError = Sites[ist]->ClockError[0];
	}

	else {
		// When several points, do L1 linear fit and predict current error.

		// Create working space (mixed use of [0...] and [1...] arrays!)
		for (ii = 0 ; ii < MAXCLKERR + 2 ; ii++) {
			aa[ii] = fvector(1, nn + 2);
		}
		xx = fvector(1, nn);
		ee = fvector(1, mm);
		
		// Fill arrays
		for (ii = 0 ; ii < mm ; ii++) {
			aa[ii][1] = Sites[ist]->ClockErrTime[ii];
			aa[ii][2] = 1.0;
			bb[ii] = Sites[ist]->ClockError[ii];
		}

		// Do the L1 fit
		l1fit(mm, nn, aa - 1, bb - 1, toler, xx, ee);
		PredictedError = xx[2] + MasterClockNow * xx[1];

		// Free the allocated storage
		free_fvector(ee, 1, mm);
		free_fvector(xx, 1, nn);

		for (ii = 0 ; ii < MAXCLKERR + 2 ; ii++) {
			free_fvector(aa[ii], 1, nn + 2);
		}
	}

	// Return corrected instrument time.
	sprintf(msgstr,
		"\r\nApplying %.2lf-s clock correction.\r\n",
		PredictedError);
	ErrWarn(hWndMain, msgstr, ist);

	return(InstClockToFix - PredictedError);
}

#endif
